@@ -1,4 +1,4 @@
-from __future__ import print_function
+from __future__ import annotations
 
 import collections
 import contextlib
@@ -6,49 +6,64 @@ import gzip
 import json
 import logging
 import sys
-import time
+import typing
 import zlib
-from datetime import datetime, timedelta
+from datetime import datetime, timedelta, timezone
+from http.client import responses
 from io import BytesIO
+from urllib.parse import urlsplit
 
 from tornado import httputil
 from tornado.web import RequestHandler
 
-from urllib3.packages.six import binary_type, ensure_str
-from urllib3.packages.six.moves.http_client import responses
-from urllib3.packages.six.moves.urllib.parse import urlsplit
+from urllib3.util.util import to_str
 
 log = logging.getLogger(__name__)
 
 
-class Response(object):
-    def __init__(self, body="", status="200 OK", headers=None):
+class Response:
+    def __init__(
+        self,
+        body: str | bytes | typing.Sequence[str | bytes] = "",
+        status: str = "200 OK",
+        headers: typing.Sequence[tuple[str, str | bytes]] | None = None,
+        json: typing.Any | None = None,
+    ) -> None:
         self.body = body
         self.status = status
-        self.headers = headers or [("Content-type", "text/plain")]
+        if json is not None:
+            self.headers = headers or [("Content-type", "application/json")]
+            self.body = json
+        else:
+            self.headers = headers or [("Content-type", "text/plain")]
 
-    def __call__(self, request_handler):
+    def __call__(self, request_handler: RequestHandler) -> None:
         status, reason = self.status.split(" ", 1)
         request_handler.set_status(int(status), reason)
         for header, value in self.headers:
             request_handler.add_header(header, value)
 
+        if isinstance(self.body, str):
+            request_handler.write(self.body.encode())
+        elif isinstance(self.body, bytes):
+            request_handler.write(self.body)
         # chunked
-        if isinstance(self.body, list):
+        else:
             for item in self.body:
                 if not isinstance(item, bytes):
                     item = item.encode("utf8")
                 request_handler.write(item)
                 request_handler.flush()
-        else:
-            body = self.body
-            if not isinstance(body, bytes):
-                body = body.encode("utf8")
-
-            request_handler.write(body)
 
 
-RETRY_TEST_NAMES = collections.defaultdict(int)
+RETRY_TEST_NAMES: dict[str, int] = collections.defaultdict(int)
+
+
+def request_params(request: httputil.HTTPServerRequest) -> dict[str, bytes]:
+    params = {}
+    for k, v in request.arguments.items():
+        params[k] = next(iter(v))
+    return params
 
 
 class TestingApp(RequestHandler):
@@ -61,32 +76,29 @@ class TestingApp(RequestHandler):
     method has its own conditions for success/failure.
     """
 
-    def get(self):
+    def get(self) -> None:
         """Handle GET requests"""
         self._call_method()
 
-    def post(self):
+    def post(self) -> None:
         """Handle POST requests"""
         self._call_method()
 
-    def put(self):
+    def put(self) -> None:
         """Handle PUT requests"""
         self._call_method()
 
-    def options(self):
+    def options(self) -> None:
         """Handle OPTIONS requests"""
         self._call_method()
 
-    def head(self):
+    def head(self) -> None:
         """Handle HEAD requests"""
         self._call_method()
 
-    def _call_method(self):
+    def _call_method(self) -> None:
         """Call the correct method in this class based on the incoming URI"""
         req = self.request
-        req.params = {}
-        for k, v in req.arguments.items():
-            req.params[k] = next(iter(v))
 
         path = req.path[:]
         if not path.startswith("/"):
@@ -103,60 +115,65 @@ class TestingApp(RequestHandler):
 
         resp(self)
 
-    def index(self, _request):
+    def index(self, _request: httputil.HTTPServerRequest) -> Response:
         "Render simple message"
         return Response("Dummy server!")
 
-    def certificate(self, request):
+    def certificate(self, request: httputil.HTTPServerRequest) -> Response:
         """Return the requester's certificate."""
         cert = request.get_ssl_certificate()
-        subject = dict()
+        assert isinstance(cert, dict)
+        subject = {}
         if cert is not None:
-            subject = dict((k, v) for (k, v) in [y for z in cert["subject"] for y in z])
+            subject = {k: v for (k, v) in [y for z in cert["subject"] for y in z]}
         return Response(json.dumps(subject))
 
-    def alpn_protocol(self, request):
+    def alpn_protocol(self, request: httputil.HTTPServerRequest) -> Response:
         """Return the selected ALPN protocol."""
-        proto = request.connection.stream.socket.selected_alpn_protocol()
-        return Response(proto.encode("utf8") if proto is not None else u"")
+        assert request.connection is not None
+        proto = request.connection.stream.socket.selected_alpn_protocol()  # type: ignore[attr-defined]
+        return Response(proto.encode("utf8") if proto is not None else "")
 
-    def source_address(self, request):
+    def source_address(self, request: httputil.HTTPServerRequest) -> Response:
         """Return the requester's IP address."""
-        return Response(request.remote_ip)
+        return Response(request.remote_ip)  # type: ignore[arg-type]
 
-    def set_up(self, request):
-        test_type = request.params.get("test_type")
-        test_id = request.params.get("test_id")
+    def set_up(self, request: httputil.HTTPServerRequest) -> Response:
+        params = request_params(request)
+        test_type = params.get("test_type")
+        test_id = params.get("test_id")
         if test_id:
-            print("\nNew test %s: %s" % (test_type, test_id))
+            print(f"\nNew test {test_type!r}: {test_id!r}")
         else:
-            print("\nNew test %s" % test_type)
+            print(f"\nNew test {test_type!r}")
         return Response("Dummy server is ready!")
 
-    def specific_method(self, request):
+    def specific_method(self, request: httputil.HTTPServerRequest) -> Response:
         "Confirm that the request matches the desired method type"
-        method = request.params.get("method")
-        if method and not isinstance(method, str):
-            method = method.decode("utf8")
+        params = request_params(request)
+        method = params.get("method")
+        method_str = method.decode() if method else None
 
-        if request.method != method:
+        if request.method != method_str:
             return Response(
-                "Wrong method: %s != %s" % (method, request.method),
+                f"Wrong method: {method_str} != {request.method}",
                 status="400 Bad Request",
             )
         return Response()
 
-    def upload(self, request):
+    def upload(self, request: httputil.HTTPServerRequest) -> Response:
         "Confirm that the uploaded file conforms to specification"
+        params = request_params(request)
         # FIXME: This is a huge broken mess
-        param = request.params.get("upload_param", b"myfile").decode("ascii")
-        filename = request.params.get("upload_filename", b"").decode("utf-8")
-        size = int(request.params.get("upload_size", "0"))
+        param = params.get("upload_param", b"myfile").decode("ascii")
+        filename = params.get("upload_filename", b"").decode("utf-8")
+        size = int(params.get("upload_size", "0"))
         files_ = request.files.get(param)
+        assert files_ is not None
 
         if len(files_) != 1:
             return Response(
-                "Expected 1 file for '%s', not %d" % (param, len(files_)),
+                f"Expected 1 file for '{param}', not {len(files_)}",
                 status="400 Bad Request",
             )
         file_ = files_[0]
@@ -164,80 +181,80 @@ class TestingApp(RequestHandler):
         data = file_["body"]
         if int(size) != len(data):
             return Response(
-                "Wrong size: %d != %d" % (size, len(data)), status="400 Bad Request"
+                f"Wrong size: {int(size)} != {len(data)}", status="400 Bad Request"
             )
 
         got_filename = file_["filename"]
-        if isinstance(got_filename, binary_type):
+        if isinstance(got_filename, bytes):
             got_filename = got_filename.decode("utf-8")
 
         # Tornado can leave the trailing \n in place on the filename.
         if filename != got_filename:
             return Response(
-                u"Wrong filename: %s != %s" % (filename, file_.filename),
+                f"Wrong filename: {filename} != {file_.filename}",
                 status="400 Bad Request",
             )
 
         return Response()
 
-    def redirect(self, request):
+    def redirect(self, request: httputil.HTTPServerRequest) -> Response:  # type: ignore[override]
         "Perform a redirect to ``target``"
-        target = request.params.get("target", "/")
-        status = request.params.get("status", "303 See Other")
+        params = request_params(request)
+        target = params.get("target", "/")
+        status = params.get("status", b"303 See Other").decode("latin-1")
         if len(status) == 3:
-            status = "%s Redirect" % status.decode("latin-1")
+            status = f"{status} Redirect"
 
         headers = [("Location", target)]
         return Response(status=status, headers=headers)
 
-    def not_found(self, request):
+    def not_found(self, request: httputil.HTTPServerRequest) -> Response:
         return Response("Not found", status="404 Not Found")
 
-    def multi_redirect(self, request):
+    def multi_redirect(self, request: httputil.HTTPServerRequest) -> Response:
         "Performs a redirect chain based on ``redirect_codes``"
-        codes = request.params.get("redirect_codes", b"200").decode("utf-8")
+        params = request_params(request)
+        codes = params.get("redirect_codes", b"200").decode("utf-8")
         head, tail = codes.split(",", 1) if "," in codes else (codes, None)
-        status = "{0} {1}".format(head, responses[int(head)])
+        assert head is not None
+        status = f"{head} {responses[int(head)]}"
         if not tail:
             return Response("Done redirecting", status=status)
 
-        headers = [("Location", "/multi_redirect?redirect_codes=%s" % tail)]
+        headers = [("Location", f"/multi_redirect?redirect_codes={tail}")]
         return Response(status=status, headers=headers)
 
-    def keepalive(self, request):
-        if request.params.get("close", b"0") == b"1":
+    def keepalive(self, request: httputil.HTTPServerRequest) -> Response:
+        params = request_params(request)
+        if params.get("close", b"0") == b"1":
             headers = [("Connection", "close")]
             return Response("Closing", headers=headers)
 
         headers = [("Connection", "keep-alive")]
         return Response("Keeping alive", headers=headers)
 
-    def echo_params(self, request):
-        params = sorted(
-            [(ensure_str(k), ensure_str(v)) for k, v in request.params.items()]
-        )
-        return Response(repr(params))
+    def echo_params(self, request: httputil.HTTPServerRequest) -> Response:
+        params = request_params(request)
+        echod = sorted((to_str(k), to_str(v)) for k, v in params.items())
+        return Response(repr(echod))
 
-    def sleep(self, request):
-        "Sleep for a specified amount of ``seconds``"
-        # DO NOT USE THIS, IT'S DEPRECATED.
-        # FIXME: Delete this once appengine tests are fixed to not use this handler.
-        seconds = float(request.params.get("seconds", "1"))
-        time.sleep(seconds)
-        return Response()
-
-    def echo(self, request):
+    def echo(self, request: httputil.HTTPServerRequest) -> Response:
         "Echo back the params"
         if request.method == "GET":
             return Response(request.query)
 
         return Response(request.body)
 
-    def echo_uri(self, request):
+    def echo_json(self, request: httputil.HTTPServerRequest) -> Response:
+        "Echo back the JSON"
+        return Response(json=request.body, headers=list(request.headers.items()))
+
+    def echo_uri(self, request: httputil.HTTPServerRequest) -> Response:
         "Echo back the requested URI"
+        assert request.uri is not None
         return Response(request.uri)
 
-    def encodingrequest(self, request):
+    def encodingrequest(self, request: httputil.HTTPServerRequest) -> Response:
         "Check for UA accepting gzip/deflate encoding"
         data = b"hello, world!"
         encoding = request.headers.get("Accept-Encoding", "")
@@ -255,16 +272,25 @@ class TestingApp(RequestHandler):
             data = zlib.compress(data)
         elif encoding == "garbage-gzip":
             headers = [("Content-Encoding", "gzip")]
-            data = "garbage"
+            data = b"garbage"
         elif encoding == "garbage-deflate":
             headers = [("Content-Encoding", "deflate")]
-            data = "garbage"
+            data = b"garbage"
         return Response(data, headers=headers)
 
-    def headers(self, request):
+    def headers(self, request: httputil.HTTPServerRequest) -> Response:
         return Response(json.dumps(dict(request.headers)))
 
-    def successful_retry(self, request):
+    def headers_and_params(self, request: httputil.HTTPServerRequest) -> Response:
+        params = request_params(request)
+        return Response(
+            json.dumps({"headers": dict(request.headers), "params": params})
+        )
+
+    def multi_headers(self, request: httputil.HTTPServerRequest) -> Response:
+        return Response(json.dumps({"headers": list(request.headers.get_all())}))
+
+    def successful_retry(self, request: httputil.HTTPServerRequest) -> Response:
         """Handler which will return an error and then success
 
         It's not currently very flexible as the number of retries is hard-coded.
@@ -280,10 +306,10 @@ class TestingApp(RequestHandler):
         else:
             return Response("need to keep retrying!", status="418 I'm A Teapot")
 
-    def chunked(self, request):
+    def chunked(self, request: httputil.HTTPServerRequest) -> Response:
         return Response(["123"] * 4)
 
-    def chunked_gzip(self, request):
+    def chunked_gzip(self, request: httputil.HTTPServerRequest) -> Response:
         chunks = []
         compressor = zlib.compressobj(6, zlib.DEFLATED, 16 + zlib.MAX_WBITS)
 
@@ -294,39 +320,45 @@ class TestingApp(RequestHandler):
 
         return Response(chunks, headers=[("Content-Encoding", "gzip")])
 
-    def nbytes(self, request):
-        length = int(request.params.get("length"))
+    def nbytes(self, request: httputil.HTTPServerRequest) -> Response:
+        params = request_params(request)
+        length = int(params["length"])
         data = b"1" * length
         return Response(data, headers=[("Content-Type", "application/octet-stream")])
 
-    def status(self, request):
-        status = request.params.get("status", "200 OK")
+    def status(self, request: httputil.HTTPServerRequest) -> Response:
+        params = request_params(request)
+        status = params.get("status", b"200 OK").decode("latin-1")
 
         return Response(status=status)
 
-    def retry_after(self, request):
-        if datetime.now() - self.application.last_req < timedelta(seconds=1):
-            status = request.params.get("status", b"429 Too Many Requests")
+    def retry_after(self, request: httputil.HTTPServerRequest) -> Response:
+        params = request_params(request)
+        if datetime.now() - self.application.last_req < timedelta(seconds=1):  # type: ignore[attr-defined]
+            status = params.get("status", b"429 Too Many Requests")
             return Response(
                 status=status.decode("utf-8"), headers=[("Retry-After", "1")]
             )
 
-        self.application.last_req = datetime.now()
+        self.application.last_req = datetime.now()  # type: ignore[attr-defined]
 
         return Response(status="200 OK")
 
-    def redirect_after(self, request):
+    def redirect_after(self, request: httputil.HTTPServerRequest) -> Response:
         "Perform a redirect to ``target``"
-        date = request.params.get("date")
+        params = request_params(request)
+        date = params.get("date")
         if date:
             retry_after = str(
-                httputil.format_timestamp(datetime.utcfromtimestamp(float(date)))
+                httputil.format_timestamp(
+                    datetime.fromtimestamp(float(date), tz=timezone.utc)
+                )
             )
         else:
             retry_after = "1"
-        target = request.params.get("target", "/")
+        target = params.get("target", "/")
         headers = [("Location", target), ("Retry-After", retry_after)]
         return Response(status="303 See Other", headers=headers)
 
-    def shutdown(self, request):
+    def shutdown(self, request: httputil.HTTPServerRequest) -> typing.NoReturn:
         sys.exit()
