@@ -1,10 +1,12 @@
+from __future__ import annotations
+
 import platform
 import select
 import socket
 import ssl
-import sys
+import typing
+from unittest import mock
 
-import mock
 import pytest
 
 from dummyserver.server import DEFAULT_CA, DEFAULT_CERTS
@@ -12,33 +14,36 @@ from dummyserver.testcase import SocketDummyServerTestCase, consume_socket
 from urllib3.util import ssl_
 from urllib3.util.ssltransport import SSLTransport
 
+if typing.TYPE_CHECKING:
+    from typing_extensions import Literal
+
 # consume_socket can iterate forever, we add timeouts to prevent halting.
 PER_TEST_TIMEOUT = 60
 
 
-def server_client_ssl_contexts():
+def server_client_ssl_contexts() -> tuple[ssl.SSLContext, ssl.SSLContext]:
     if hasattr(ssl, "PROTOCOL_TLS_SERVER"):
         server_context = ssl.SSLContext(ssl.PROTOCOL_TLS_SERVER)
-    else:
-        # python 2.7 and 3.5 workaround.
-        # PROTOCOL_TLS_SERVER was added in 3.6
-        server_context = ssl.SSLContext(ssl.PROTOCOL_TLS)
     server_context.load_cert_chain(DEFAULT_CERTS["certfile"], DEFAULT_CERTS["keyfile"])
 
     if hasattr(ssl, "PROTOCOL_TLS_CLIENT"):
         client_context = ssl.SSLContext(ssl.PROTOCOL_TLS_CLIENT)
-    else:
-        # python 2.7 and 3.5 workaround.
-        # PROTOCOL_TLS_SERVER was added in 3.6
-        client_context = ssl.SSLContext(ssl.PROTOCOL_TLS)
-        client_context.verify_mode = ssl.CERT_REQUIRED
-        client_context.check_hostname = True
 
     client_context.load_verify_locations(DEFAULT_CA)
     return server_context, client_context
 
 
-def sample_request(binary=True):
+@typing.overload
+def sample_request(binary: Literal[True] = ...) -> bytes:
+    ...
+
+
+@typing.overload
+def sample_request(binary: Literal[False]) -> str:
+    ...
+
+
+def sample_request(binary: bool = True) -> bytes | str:
     request = (
         b"GET http://www.testing.com/ HTTP/1.1\r\n"
         b"Host: www.testing.com\r\n"
@@ -48,25 +53,43 @@ def sample_request(binary=True):
     return request if binary else request.decode("utf-8")
 
 
-def validate_request(provided_request, binary=True):
+def validate_request(
+    provided_request: bytearray, binary: Literal[False, True] = True
+) -> None:
     assert provided_request is not None
     expected_request = sample_request(binary)
     assert provided_request == expected_request
 
 
-def sample_response(binary=True):
+@typing.overload
+def sample_response(binary: Literal[True] = ...) -> bytes:
+    ...
+
+
+@typing.overload
+def sample_response(binary: Literal[False]) -> str:
+    ...
+
+
+@typing.overload
+def sample_response(binary: bool = ...) -> bytes | str:
+    ...
+
+
+def sample_response(binary: bool = True) -> bytes | str:
     response = b"HTTP/1.1 200 OK\r\nContent-Length: 0\r\n\r\n"
     return response if binary else response.decode("utf-8")
 
 
-def validate_response(provided_response, binary=True):
+def validate_response(
+    provided_response: bytes | bytearray | str, binary: bool = True
+) -> None:
     assert provided_response is not None
     expected_response = sample_response(binary)
     assert provided_response == expected_response
 
 
-def validate_peercert(ssl_socket):
-
+def validate_peercert(ssl_socket: SSLTransport) -> None:
     binary_cert = ssl_socket.getpeercert(binary_form=True)
     assert type(binary_cert) == bytes
     assert len(binary_cert) > 0
@@ -77,7 +100,6 @@ def validate_peercert(ssl_socket):
     assert cert["serialNumber"] != ""
 
 
-@pytest.mark.skipif(sys.version_info < (3, 5), reason="requires python3.5 or higher")
 class SingleTLSLayerTestCase(SocketDummyServerTestCase):
     """
     Uses the SocketDummyServer to validate a single TLS layer can be
@@ -85,22 +107,27 @@ class SingleTLSLayerTestCase(SocketDummyServerTestCase):
     """
 
     @classmethod
-    def setup_class(cls):
+    def setup_class(cls) -> None:
         cls.server_context, cls.client_context = server_client_ssl_contexts()
 
-    def start_dummy_server(self, handler=None):
-        def socket_handler(listener):
+    def start_dummy_server(
+        self, handler: typing.Callable[[socket.socket], None] | None = None
+    ) -> None:
+        def socket_handler(listener: socket.socket) -> None:
             sock = listener.accept()[0]
-            with self.server_context.wrap_socket(sock, server_side=True) as ssock:
-                request = consume_socket(ssock)
-                validate_request(request)
-                ssock.send(sample_response())
+            try:
+                with self.server_context.wrap_socket(sock, server_side=True) as ssock:
+                    request = consume_socket(ssock)
+                    validate_request(request)
+                    ssock.send(sample_response())
+            except (ConnectionAbortedError, ConnectionResetError):
+                return
 
         chosen_handler = handler if handler else socket_handler
         self._start_server(chosen_handler)
 
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_start_closed_socket(self):
+    def test_start_closed_socket(self) -> None:
         """Errors generated from an unconnected socket should bubble up."""
         sock = socket.socket(socket.AF_INET)
         context = ssl.create_default_context()
@@ -109,7 +136,7 @@ class SingleTLSLayerTestCase(SocketDummyServerTestCase):
             SSLTransport(sock, context)
 
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_close_after_handshake(self):
+    def test_close_after_handshake(self) -> None:
         """Socket errors should be bubbled up"""
         self.start_dummy_server()
 
@@ -122,7 +149,7 @@ class SingleTLSLayerTestCase(SocketDummyServerTestCase):
                 ssock.send(b"blaaargh")
 
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_wrap_existing_socket(self):
+    def test_wrap_existing_socket(self) -> None:
         """Validates a single TLS layer can be established."""
         self.start_dummy_server()
 
@@ -136,7 +163,7 @@ class SingleTLSLayerTestCase(SocketDummyServerTestCase):
             validate_response(response)
 
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_unbuffered_text_makefile(self):
+    def test_unbuffered_text_makefile(self) -> None:
         self.start_dummy_server()
 
         sock = socket.create_connection((self.host, self.port))
@@ -150,13 +177,13 @@ class SingleTLSLayerTestCase(SocketDummyServerTestCase):
             validate_response(response)
 
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_unwrap_existing_socket(self):
+    def test_unwrap_existing_socket(self) -> None:
         """
         Validates we can break up the TLS layer
         A full request/response is sent over TLS, and later over plain text.
         """
 
-        def shutdown_handler(listener):
+        def shutdown_handler(listener: socket.socket) -> None:
             sock = listener.accept()[0]
             ssl_sock = self.server_context.wrap_socket(sock, server_side=True)
 
@@ -186,7 +213,7 @@ class SingleTLSLayerTestCase(SocketDummyServerTestCase):
         validate_response(response)
 
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_ssl_object_attributes(self):
+    def test_ssl_object_attributes(self) -> None:
         """Ensures common ssl attributes are exposed"""
         self.start_dummy_server()
 
@@ -202,8 +229,11 @@ class SingleTLSLayerTestCase(SocketDummyServerTestCase):
             assert ssock.selected_npn_protocol() is None
 
             shared_ciphers = ssock.shared_ciphers()
-            assert type(shared_ciphers) == list
-            assert len(shared_ciphers) > 0
+            # SSLContext.shared_ciphers() changed behavior completely in a patch version.
+            # See: https://github.com/python/cpython/issues/96931
+            assert shared_ciphers is None or (
+                type(shared_ciphers) is list and len(shared_ciphers) > 0
+            )
 
             assert ssock.compression() is None
 
@@ -214,7 +244,7 @@ class SingleTLSLayerTestCase(SocketDummyServerTestCase):
             validate_response(response)
 
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_socket_object_attributes(self):
+    def test_socket_object_attributes(self) -> None:
         """Ensures common socket attributes are exposed"""
         self.start_dummy_server()
 
@@ -239,20 +269,22 @@ class SocketProxyDummyServer(SocketDummyServerTestCase):
     socket.
     """
 
-    def __init__(self, destination_server_host, destination_server_port):
+    def __init__(
+        self, destination_server_host: str, destination_server_port: int
+    ) -> None:
         self.destination_server_host = destination_server_host
         self.destination_server_port = destination_server_port
-        self.server_context, self.client_context = server_client_ssl_contexts()
+        self.server_ctx, _ = server_client_ssl_contexts()
 
-    def start_proxy_handler(self):
+    def start_proxy_handler(self) -> None:
         """
         Socket handler for the proxy. Terminates the first TLS layer and tunnels
         any bytes needed for client <-> server communicatin.
         """
 
-        def proxy_handler(listener):
+        def proxy_handler(listener: socket.socket) -> None:
             sock = listener.accept()[0]
-            with self.server_context.wrap_socket(sock, server_side=True) as client_sock:
+            with self.server_ctx.wrap_socket(sock, server_side=True) as client_sock:
                 upstream_sock = socket.create_connection(
                     (self.destination_server_host, self.destination_server_port)
                 )
@@ -262,7 +294,12 @@ class SocketProxyDummyServer(SocketDummyServerTestCase):
 
         self._start_server(proxy_handler)
 
-    def _read_write_loop(self, client_sock, server_sock, chunks=65536):
+    def _read_write_loop(
+        self,
+        client_sock: socket.socket,
+        server_sock: socket.socket,
+        chunks: int = 65536,
+    ) -> None:
         inputs = [client_sock, server_sock]
         output = [client_sock, server_sock]
 
@@ -270,7 +307,7 @@ class SocketProxyDummyServer(SocketDummyServerTestCase):
             readable, writable, exception = select.select(inputs, output, inputs)
 
             if exception:
-                # Error ocurred with either of the sockets, time to
+                # Error occurred with either of the sockets, time to
                 # wrap up, parent func will close sockets.
                 break
 
@@ -283,7 +320,7 @@ class SocketProxyDummyServer(SocketDummyServerTestCase):
                     read_socket = server_sock
                     write_socket = client_sock
 
-                # Ensure buffer is not full before writting
+                # Ensure buffer is not full before writing
                 if write_socket in writable:
                     try:
                         b = read_socket.recv(chunks)
@@ -299,7 +336,6 @@ class SocketProxyDummyServer(SocketDummyServerTestCase):
                         return
 
 
-@pytest.mark.skipif(sys.version_info < (3, 5), reason="requires python3.5 or higher")
 class TlsInTlsTestCase(SocketDummyServerTestCase):
     """
     Creates a TLS in TLS tunnel by chaining a 'SocketProxyDummyServer' and a
@@ -311,41 +347,44 @@ class TlsInTlsTestCase(SocketDummyServerTestCase):
     """
 
     @classmethod
-    def setup_class(cls):
+    def setup_class(cls) -> None:
         cls.server_context, cls.client_context = server_client_ssl_contexts()
 
     @classmethod
-    def start_proxy_server(cls):
+    def start_proxy_server(cls) -> None:
         # Proxy server will handle the first TLS connection and create a
         # connection to the destination server.
         cls.proxy_server = SocketProxyDummyServer(cls.host, cls.port)
         cls.proxy_server.start_proxy_handler()
 
     @classmethod
-    def teardown_class(cls):
+    def teardown_class(cls) -> None:
         if hasattr(cls, "proxy_server"):
             cls.proxy_server.teardown_class()
-        super(TlsInTlsTestCase, cls).teardown_class()
+        super().teardown_class()
 
     @classmethod
-    def start_destination_server(cls):
+    def start_destination_server(cls) -> None:
         """
         Socket handler for the destination_server. Terminates the second TLS
         layer and send a basic HTTP response.
         """
 
-        def socket_handler(listener):
+        def socket_handler(listener: socket.socket) -> None:
             sock = listener.accept()[0]
-            with cls.server_context.wrap_socket(sock, server_side=True) as ssock:
-                request = consume_socket(ssock)
-                validate_request(request)
-                ssock.send(sample_response())
+            try:
+                with cls.server_context.wrap_socket(sock, server_side=True) as ssock:
+                    request = consume_socket(ssock)
+                    validate_request(request)
+                    ssock.send(sample_response())
+            except (ssl.SSLEOFError, ssl.SSLZeroReturnError, OSError):
+                return
             sock.close()
 
         cls._start_server(socket_handler)
 
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_tls_in_tls_tunnel(self):
+    def test_tls_in_tls_tunnel(self) -> None:
         """
         Basic communication over the TLS in TLS tunnel.
         """
@@ -367,7 +406,7 @@ class TlsInTlsTestCase(SocketDummyServerTestCase):
                 validate_response(response)
 
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_wrong_sni_hint(self):
+    def test_wrong_sni_hint(self) -> None:
         """
         Provides a wrong sni hint to validate an exception is thrown.
         """
@@ -380,17 +419,14 @@ class TlsInTlsTestCase(SocketDummyServerTestCase):
         with self.client_context.wrap_socket(
             sock, server_hostname="localhost"
         ) as proxy_sock:
-            with pytest.raises(Exception) as e:
+            with pytest.raises(ssl.SSLCertVerificationError):
                 SSLTransport(
                     proxy_sock, self.client_context, server_hostname="veryverywrong"
                 )
-            # ssl.CertificateError is a child of ValueError in python3.6 or
-            # before. After python3.7 it's a child of SSLError
-            assert e.type in [ssl.SSLError, ssl.CertificateError]
 
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
     @pytest.mark.parametrize("buffering", [None, 0])
-    def test_tls_in_tls_makefile_raw_rw_binary(self, buffering):
+    def test_tls_in_tls_makefile_raw_rw_binary(self, buffering: int | None) -> None:
         """
         Uses makefile with read, write and binary modes without buffering.
         """
@@ -406,13 +442,12 @@ class TlsInTlsTestCase(SocketDummyServerTestCase):
             with SSLTransport(
                 proxy_sock, self.client_context, server_hostname="localhost"
             ) as destination_sock:
-
                 file = destination_sock.makefile("rwb", buffering)
-                file.write(sample_request())
+                file.write(sample_request())  # type: ignore[call-overload]
                 file.flush()
 
                 response = bytearray(65536)
-                wrote = file.readinto(response)
+                wrote = file.readinto(response)  # type: ignore[union-attr]
                 assert wrote is not None
                 # Allocated response is bigger than the actual response, we
                 # rtrim remaining x00 bytes.
@@ -425,7 +460,7 @@ class TlsInTlsTestCase(SocketDummyServerTestCase):
         reason="Skipping windows due to text makefile support",
     )
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_tls_in_tls_makefile_rw_text(self):
+    def test_tls_in_tls_makefile_rw_text(self) -> None:
         """
         Creates a separate buffer for reading and writing using text mode and
         utf-8 encoding.
@@ -442,22 +477,23 @@ class TlsInTlsTestCase(SocketDummyServerTestCase):
             with SSLTransport(
                 proxy_sock, self.client_context, server_hostname="localhost"
             ) as destination_sock:
-
                 read = destination_sock.makefile("r", encoding="utf-8")
                 write = destination_sock.makefile("w", encoding="utf-8")
 
-                write.write(sample_request(binary=False))
+                write.write(sample_request(binary=False))  # type: ignore[arg-type, call-overload]
                 write.flush()
 
                 response = read.read()
+                assert isinstance(response, str)
                 if "\r" not in response:
                     # Carriage return will be removed when reading as a file on
                     # some platforms.  We add it before the comparison.
+                    assert isinstance(response, str)
                     response = response.replace("\n", "\r\n")
                 validate_response(response, binary=False)
 
     @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_tls_in_tls_recv_into_sendall(self):
+    def test_tls_in_tls_recv_into_sendall(self) -> None:
         """
         Valides recv_into and sendall also work as expected. Other tests are
         using recv/send.
@@ -474,39 +510,15 @@ class TlsInTlsTestCase(SocketDummyServerTestCase):
             with SSLTransport(
                 proxy_sock, self.client_context, server_hostname="localhost"
             ) as destination_sock:
-
                 destination_sock.sendall(sample_request())
                 response = bytearray(65536)
                 destination_sock.recv_into(response)
                 str_response = response.decode("utf-8").rstrip("\x00")
                 validate_response(str_response, binary=False)
 
-    @pytest.mark.timeout(PER_TEST_TIMEOUT)
-    def test_tls_in_tls_recv_into_unbuffered(self):
-        """
-        Valides recv_into without a preallocated buffer.
-        """
-        self.start_destination_server()
-        self.start_proxy_server()
 
-        sock = socket.create_connection(
-            (self.proxy_server.host, self.proxy_server.port)
-        )
-        with self.client_context.wrap_socket(
-            sock, server_hostname="localhost"
-        ) as proxy_sock:
-            with SSLTransport(
-                proxy_sock, self.client_context, server_hostname="localhost"
-            ) as destination_sock:
-
-                destination_sock.send(sample_request())
-                response = destination_sock.recv_into(None)
-                validate_response(response)
-
-
-@pytest.mark.skipif(sys.version_info < (3, 5), reason="requires python3.5 or higher")
-class TestSSLTransportWithMock(object):
-    def test_constructor_params(self):
+class TestSSLTransportWithMock:
+    def test_constructor_params(self) -> None:
         server_hostname = "example-domain.com"
         sock = mock.Mock()
         context = mock.create_autospec(ssl_.SSLContext)
@@ -517,3 +529,44 @@ class TestSSLTransportWithMock(object):
             mock.ANY, mock.ANY, server_hostname=server_hostname
         )
         assert not ssl_transport.suppress_ragged_eofs
+
+    def test_various_flags_errors(self) -> None:
+        server_hostname = "example-domain.com"
+        sock = mock.Mock()
+        context = mock.create_autospec(ssl_.SSLContext)
+        ssl_transport = SSLTransport(
+            sock, context, server_hostname=server_hostname, suppress_ragged_eofs=False
+        )
+        with pytest.raises(ValueError):
+            ssl_transport.recv(flags=1)
+
+        with pytest.raises(ValueError):
+            ssl_transport.recv_into(bytearray(), flags=1)
+
+        with pytest.raises(ValueError):
+            ssl_transport.sendall(bytearray(), flags=1)
+
+        with pytest.raises(ValueError):
+            ssl_transport.send(None, flags=1)  # type: ignore[arg-type]
+
+    def test_makefile_wrong_mode_error(self) -> None:
+        server_hostname = "example-domain.com"
+        sock = mock.Mock()
+        context = mock.create_autospec(ssl_.SSLContext)
+        ssl_transport = SSLTransport(
+            sock, context, server_hostname=server_hostname, suppress_ragged_eofs=False
+        )
+        with pytest.raises(ValueError):
+            ssl_transport.makefile(mode="x")
+
+    def test_wrap_ssl_read_error(self) -> None:
+        server_hostname = "example-domain.com"
+        sock = mock.Mock()
+        context = mock.create_autospec(ssl_.SSLContext)
+        ssl_transport = SSLTransport(
+            sock, context, server_hostname=server_hostname, suppress_ragged_eofs=False
+        )
+        with mock.patch.object(ssl_transport, "_ssl_io_loop") as _ssl_io_loop:
+            _ssl_io_loop.side_effect = ssl.SSLError()
+            with pytest.raises(ssl.SSLError):
+                ssl_transport._wrap_ssl_read(1)
